/* Copyright 2012 Cloudseal Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.cloudseal.client.tomcat;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletResponse;

import com.cloudseal.client.saml2.*;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.Session;
import org.apache.catalina.authenticator.Constants;
import org.apache.catalina.authenticator.FormAuthenticator;
import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.deploy.LoginConfig;
import org.apache.catalina.realm.GenericPrincipal;
import org.apache.commons.io.IOUtils;
import org.apache.juli.logging.Log;
import org.apache.juli.logging.LogFactory;

import java.io.*;
import java.security.Principal;
import java.util.ArrayList;
import java.util.List;

/**
 * Adds <a href="http://www.cloudseal.com">Cloudseal</a> authentication and Single Sign On to Tomcat. Although the
 * valve has some Cloudseal specific extensions (user attributes, authorities and single logout) it should work
 * with any SAML2 IDP which supports the Web browser SSO profile and redirect/post binding, although it has only
 * been tested with Cloudseal. The valve should be configured for each context e.g.
 *
 * <p>
 * &lt;Context&gt;<br />
 * <span style="margin-left:20px">&lt;Valve className="com.cloudseal.client.tomcat.CloudsealAuthenticator"</span><br />
 * <span style="margin-left:40px">keystore="/WEB-INF/keystore.jks" keystorePassword="password" key="cloudseal"</span><br />
 * <span style="margin-left:40px">keyPassword="password" idp="/WEB-INF/idp.xml" appId="test"&gt;</span><br />
 * <span style="margin-left:20px">&lt;/Valve&gt;</span><br />
 * &lt;/Context&gt;
 * </p>
 *
 * <p>Sample web.xml: </p>
 * <p>
 *  &lt;security-constraint&gt;<br />
 * <span style="margin-left:20px">&lt;web-resource-collection&gt;</span><br />
 * <span style="margin-left:40px">&lt;web-resource-name&gt;protected&lt;/web-resource-name&gt;</span><br />
 * <span style="margin-left:40px">&lt;url-pattern&gt;/protected.html&lt;/url-pattern&gt;</span><br />
 * <span style="margin-left:20px">&lt;/web-resource-collection&gt;</span><br />
 * <span style="margin-left:20px">&lt;auth-constraint&gt;</span><br />
 * <span style="margin-left:40px">&lt;role-name&gt;ROLE_ADMIN&lt;/role-name&gt;</span><br />
 * <span style="margin-left:20px">&lt;/auth-constraint&gt;</span><br />
 * &lt;/security-constraint&gt;<br />
 * &lt;security-role&gt;<br />
 * <span style="margin-left:20px">&lt;role-name&gt;ROLE_ADMIN&lt;/role-name&gt;</span><br />
 * &lt;/security-role&gt;
 * </p>
 *
 * <p>Full documentation can be found <a href="https://cloudseal.jira.com/wiki/display/WIKI/Tomcat+valve">here</a></p>
 */
public class CloudsealAuthenticator extends FormAuthenticator {

    // If the developer specifies an auth-method in web.xml it must be set to this value
    private static final String AUTH_METHOD = "CLOUDSEAL";

    // Note: Cloudseal specific, The Cloudseal platform will trigger a request from the user's browser
    // to this url and the valve will intercept it, perform a logout and then return a json message
    // see https://cloudseal.jira.com/wiki/display/WIKI/Cloudseal+logout+protocol
    private static final String BROWSER_INITIATED_SSO_URL = "/logout.jsonp";

    // A session key for the SAML request id
    private static final String ID = "com.cloudseal.client.tomcat.Authenticator.Id";

    // A session key for the SAML audience
    private static final String AUDIENCE = "com.cloudseal.client.tomcat.Authenticator.Audience";

    private static final Log LOG = LogFactory.getLog(CloudsealAuthenticator.class);

    // Generates a random string
    private IdentifierGenerator idGenerator;

    // Builds the authentication request
    private AuthRequestBuilder authRequestBuilder;

    // Validates the authentication response from Cloudseal
    private AuthResponseValidator responseValidator;

    // Parses the idp.xml file and extracts the urls and the public key that we
    // will use to verify the response from Cloudseal is authentic
    private IdpXmlParser idpXmlParser;

    // location of idp.xml
    private String idp;

    // location of the keystore
    private String keystore;

    // keystore password
    private String keystorePassword;

    // keystore key (alias) name
    private String key;

    // key password
    private String keyPassword;

    // optional application id used to map Cloudseal roles to J2EE roles
    private String appId;

    public CloudsealAuthenticator() throws Exception {
        setAlwaysUseSession(true);
    }

    @Override
    protected String getAuthMethod() {
        return AUTH_METHOD;
    }

    @Override
    protected synchronized void initInternal() throws LifecycleException {
        super.initInternal();
        LOG.info("Initializing");
        // Validate that all required properties are set
        validateConfigParam(idp, "idp");
        validateConfigParam(keystore, "keystore");
        validateConfigParam(keystorePassword, "keystorePassword");
        validateConfigParam(key, "key");
        validateConfigParam(keyPassword, "keyPassword");

        try {
            // AuthRequestContext is simply a container class that allows us to pass data between the Cloudseal objects
            AuthRequestContext authRequestContext = new AuthRequestContext();
            authRequestContext.setKeyName(key);
            authRequestContext.setKeyPassword(keyPassword);
            authRequestContext.setKeystore(getResource(keystore));
            authRequestContext.setKeystorePassword(keystorePassword);
            authRequestContext.setProviderName(appId);

            // Create and initialize the Cloudseal objects, the if null tests are there to aid unit testing
            // as we also have package visible setters for these objects (which could be mocks)
            if (authRequestBuilder == null)
                authRequestBuilder = new AuthRequestBuilderImpl();

            authRequestBuilder.init(authRequestContext);

            if (responseValidator == null)
                responseValidator = new AuthResponseValidatorImpl();

            if (idpXmlParser == null)
                idpXmlParser = new IdpXmlParser();

            idpXmlParser.parse(getResource(idp));

            if (idGenerator == null)
                idGenerator = new IdentifierGenerator();
        } catch (Exception e) {
            throw new LifecycleException("Failed to initialize", e);
        }
    }

    @Override
    public void invoke(Request request, Response response) throws IOException, ServletException {
        LOG.info("invoke()");

        // single log out request? Note: this is Cloudseal specific
        if (request.getRequestURL().toString().endsWith(BROWSER_INITIATED_SSO_URL)) {
            LOG.info("browser initiated single log out");
            logout(request);
            // Send the jsonp response back to Cloudseal
            String callback = request.getParameter("callback");
            String jsonp = String.format("%s({\"status\":\"success\"})", callback);
            response.getWriter().print(jsonp);
            response.getWriter().flush();
            return;
        }

        if (context.getLoginConfig() != null) {
            String authMethod = context.getLoginConfig().getAuthMethod();
            /*
             * Without registering the authenticator and its auth method in the container,
             * it won't even invoke the valve if unknown auth method encountered in app's web.xml.
             * If login-config is omitted then by default auth method is NONE.
             */
            if (AUTH_METHOD.equals(authMethod) || "NONE".equals(authMethod)) {
                super.invoke(request, response);
                return;
            }
        }

        getNext().invoke(request, response);
    }

    /**
     * Authenticate the user making this request, based on the specified
     * login configuration.  Return <code>true</code> if any specified
     * constraint has been satisfied, or <code>false</code> if we have
     * created a response challenge already.
     *
     * Note: most of the code here is lifted from the FormAuthenticator
     * so there is scope to refactor
     *
     * @param request   Request we are processing
     * @param response  Response we are creating
     * @param config    Login configuration describing how authentication
     *                  should be performed
     *
     * @exception IOException if an input/output error occurs
     */
    @Override
    public boolean authenticate(Request request,
                                HttpServletResponse response,
                                LoginConfig config)
            throws IOException {

        // References to objects we will need later
        Session session = null;

        // Have we already authenticated someone?
        Principal principal = request.getUserPrincipal();
        String ssoId = (String) request.getNote(Constants.REQ_SSOID_NOTE);
        if (principal != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Already authenticated '" +
                        principal.getName() + "'");
            }
            // Associate the session with any existing SSO session
            if (ssoId != null) {
                associate(ssoId, request.getSessionInternal(true));
            }
            return (true);
        }  // if principal != null

        // Is there an SSO session against which we can try to reauthenticate?
        if (ssoId != null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("SSO Id " + ssoId + " set; attempting " +
                        "reauthentication");
            }
            // Try to reauthenticate using data cached by SSO.  If this fails,
            // either the original SSO logon was of DIGEST or SSL (which
            // we can't reauthenticate ourselves because there is no
            // cached username and password), or the realm denied
            // the user's reauthentication for some reason.
            // In either case we have to prompt the user for a logon */
            if (reauthenticateFromSSO(ssoId, request)) {
                return true;
            }
        } // id ssoId != null

        // Is this the re-submit of the original request URI after successful
        // authentication?  If so, forward the *original* request instead.
        if (matchRequest(request)) {
            session = request.getSessionInternal(true);
            if (LOG.isDebugEnabled()) {
                LOG.debug("Restore request from session '"
                        + session.getIdInternal()
                        + "'");
            }

            // Cloudseal principal includes additional fields e.g. firstName, lastName etc
            // which are passed from the Cloudseal IDP, however it will happily work with another
            // SAML2 IDP, but the Cloudseal specific fields (which are SAML2 attributes)
            // will be null
            CloudsealPrincipal cloudsealPrincipal = (CloudsealPrincipal)
                    session.getNote(Constants.FORM_PRINCIPAL_NOTE);

            // Maps the Cloudseal roles to tomcat
            List<String> roles = new ArrayList<String>(cloudsealPrincipal.getRoles());
            GenericPrincipal genericPrincipal = new GenericPrincipal(cloudsealPrincipal.getUsername(),
                    null, roles, cloudsealPrincipal);

            // authenticate and save to session (this will create new session)
            register(request, response, genericPrincipal, getAuthMethod(), genericPrincipal.getName(), null);

            if (restoreRequest(request, session)) {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Proceed to restored request");
                }
                return (true);
            } else {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Restore of original request failed");
                }
                response.sendError(HttpServletResponse.SC_BAD_REQUEST);
                return (false);
            }
        }  // if matchRequest

        // is this a response from the IDP
        boolean samlResponse = request.getParameter("SAMLResponse") != null;

        // No -- Save this request and redirect to the IDP
        if (! samlResponse) {
            session = request.getSessionInternal(true);
            if (LOG.isDebugEnabled()) {
                LOG.debug("Save request in session '" + session.getIdInternal() + "'");
            }
            try {
                saveRequest(request, session);
            } catch (IOException ioe) {
                LOG.debug("Request body too big to save during authentication");
                response.sendError(HttpServletResponse.SC_FORBIDDEN, "Response body too big");
                return (false);
            }
            redirectToIdp(request, response);
            return (false);
        } // if ! samlResponse

        // Yes -- validate the specified credentials
        // and redirect to the error page if they are not correct
        session = request.getSessionInternal(false);
        if (session == null) {
            if (LOG.isDebugEnabled()) {
                LOG.debug("User took so long to log on the session expired");
            }
            response.sendError(HttpServletResponse.SC_REQUEST_TIMEOUT, "Session expired");
            return (false);
        }

        // SAML2 concepts, when we redirect the user to the IDP we pass these attributes
        // and we expect the response to match
        String expectedInResponseTo = (String) session.getSession().getAttribute(ID);
        String expectedAudience = (String) session.getSession().getAttribute(AUDIENCE);
        try {
            principal = responseValidator.validateResponse(idpXmlParser.getIdpPublicKey(),
                    request.getParameter("SAMLResponse"), expectedInResponseTo, expectedAudience);
            LOG.debug("Authentication of '" + principal.getName() + "' was successful");
            session.getSession().removeAttribute(ID);
            session.getSession().removeAttribute(AUDIENCE);
        } catch (VerificationException ex) {
            LOG.error("Cloudseal verification failed: ", ex);
            forwardToErrorPage(request, response, config);
            return (false);
        }

        // Save the authenticated Principal in our session
        // Note we need to use Constants.FORM_PRICIPAL_NOTE so we can reuse the matchRequest()
        // method in FormAuthenticator
        session.setNote(Constants.FORM_PRINCIPAL_NOTE, principal);

        // Redirect the user to the original request URI (which will cause
        // the original request to be restored)
        String requestURI = savedRequestURL(session);
        if (LOG.isDebugEnabled()) {
            LOG.debug("Redirecting to original '" + requestURI + "'");
        }
        if (requestURI == null) {
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Cloudseal authenticator");
        } else {
            response.sendRedirect(response.encodeRedirectURL(requestURI));
        }
        return (false);
    }

    /**
     * Send an authentication request to Cloudseal
     *
     * @param request
     * @param response
     * @throws IOException
     */
    void redirectToIdp(Request request, HttpServletResponse response) throws IOException {
        try {
            String spIssuer = buildSpIssuer(request);
            String id = idGenerator.generateIdentifier();
            String acsUrl = buildAcsUrl(request);
            request.getSession().setAttribute(ID, id);
            request.getSession().setAttribute(AUDIENCE, spIssuer);
            String samlAuthRequest = authRequestBuilder.generateSamlAuthRequest(spIssuer, acsUrl, spIssuer, id);
            String redirectUrl = idpXmlParser.getSsoUrl() + "?" + samlAuthRequest;
            response.sendRedirect(redirectUrl);
        } catch (Exception e) {
            LOG.error(e.getMessage(), e);
            sendForbiddenError(response, e.getMessage());
        }
    }

    /**
     * Attempt to locate a resource. By default it will look for the named resource using the ServletContext, however
     * if the classpath: prefix is used it will look for the resource on the classpath (which will be $CATALINA_HOME/lib
     * as this is where this class will reside)
     *
     * @param filename
     * @return
     * @throws IOException
     */
    byte[] getResource(String filename) throws IOException {
        ServletContext servletContext = this.context.getServletContext();
        InputStream is;
        if (filename.startsWith("classpath:")) {
            String path = filename.split(":")[1];
            is = this.getClass().getClassLoader().getResourceAsStream(path);
        } else {
            is = servletContext.getResource(filename).openStream();
        }
        try {
            return IOUtils.toByteArray(is);
        } finally {
            IOUtils.closeQuietly(is);
        }
    }

    /**
     * All SAML2 requests need an Issuer element, we use the hostname entered by the user
     *
     * @param request
     * @return
     */
    String buildSpIssuer(Request request) {
        String host = request.getHeader("Host");
        String protocol = "http://";
        if (host != null && host.length() > 0) {
            return protocol + host + "/saml/sp";
        } else {
            return protocol + "www.mycorp.com" + "/saml/sp";
        }
    }

    /**
     * Work out where to send the Cloudseal response to. We use the current url because we can be sure it will
     * hit the valve.
     *
     * @param request
     * @return
     */
    String buildAcsUrl(Request request) {
        return request.getRequestURL().toString();
    }

    void validateConfigParam(String param, String paramName) throws LifecycleException {
        if (param == null || param.length() < 1) {
            throw new LifecycleException("Missing configuration parameter: " + paramName);
        }
    }

    void sendForbiddenError(HttpServletResponse response, String error) throws IOException {
        response.sendError(HttpServletResponse.SC_FORBIDDEN, error);
    }

    /**
     * Location of the idp.xml file relative to the ServletContext e.g /WEB-INF/idp.xml
     *
     * @param idp
     */
    public void setIdp(String idp) {
        this.idp = idp;
    }

    /**
     * Location of the keystore relative to the ServletContext e.g /WEB-INF/keystore.jks
     * @param keystore
     */
    public void setKeystore(String keystore) {
        this.keystore = keystore;
    }

    /**
     * Keystore password
     *
     * @param keystorePassword
     */
    public void setKeystorePassword(String keystorePassword) {
        this.keystorePassword = keystorePassword;
    }

    /**
     * Keystore key (alias) name, used to sign authentication requests
     *
     * @param key
     */
    public void setKey(String key) {
        this.key = key;
    }

    /**
     * Password for the key
     *
     * @param keyPassword
     */
    public void setKeyPassword(String keyPassword) {
        this.keyPassword = keyPassword;
    }

    /**
     * Application id as defined in the Cloudseal admin console. Needed for single logout and Cloudseal role based
     * access control
     *
     * @param appId
     */
    public void setAppId(String appId) {
        this.appId = appId;
    }

    /*
     * Allows for easier unit testing
     */
    void setIdGenerator(IdentifierGenerator idGenerator) {
        this.idGenerator = idGenerator;
    }

    void setAuthRequestBuilder(AuthRequestBuilderImpl authRequestBuilder) {
        this.authRequestBuilder = authRequestBuilder;
    }

    void setResponseValidator(AuthResponseValidatorImpl responseValidator) {
        this.responseValidator = responseValidator;
    }

    void setIdpXmlParser(IdpXmlParser idpXmlParser) {
        this.idpXmlParser = idpXmlParser;
    }

}
